from torch.utils.data import Dataset, IterableDataset


class OrderDatasetLocalIterator(IterableDataset):
    """
    TODO: https://pytorch.org/docs/stable/data.html#torch.utils.data.IterableDataset
    """
    def __init__(self):
        pass

    def __len__(self):
        pass

    def __getitem__(self, idx):
        pass
